"""
twitch.ext.oauth

An extension module to facilitate the authorization

:copyright: (c) 2024-present Snifo
:license: MIT, see LICENSE for more details.
"""

from .oauth import DeviceAuthFlow
from .enums import Scopes
